@@ -16,8 +16,8 @@
 import json
 import logging
 import pprint
-import Queue
 import select
+import six.moves
 import threading
 import time
 
@@ -29,6 +29,11 @@ CONSUMING = 'consuming'
 DEAD = 'dead'
 DISCONNECTED = 'disconnected'
 IDLE = 'idle'
+UPDATE_ALLOWED_KEYS = ['description', 'submit-type',
+                       'contributor-agreements', 'signed-off-by',
+                       'content-merge', 'change-id',
+                       'project-state',
+                       'max-object-size-limit']
 
 
 class GerritWatcher(threading.Thread):
@@ -182,7 +187,7 @@ class Gerrit(object):
         self.installed_plugins = None
 
     def startWatching(self, connection_attempts=-1, retry_delay=5):
-        self.event_queue = Queue.Queue()
+        self.event_queue = six.moves.queue.Queue()
         watcher = GerritWatcher(self,
                                 connection_attempts=connection_attempts,
                                 retry_delay=retry_delay)
@@ -206,23 +211,87 @@ class Gerrit(object):
         out, err = self._ssh(cmd)
         return err
 
-    def createProject(self, project, require_change_id=True, empty_repo=False):
+    def listMembers(self, group, recursive=False):
+        cmd = 'gerrit ls-members'
+        if recursive:
+            cmd = '%s --recursive' % cmd
+        cmd = '%s "%s"' % (cmd, group)
+        out, err = self._ssh(cmd)
+        ret = []
+        rows = out.split('\n')
+        if len(rows) > 1:
+            keys = rows[0].split('\t')
+            for row in rows[1:]:
+                ret.append(dict(zip(keys, row.split('\t'))))
+        return ret
+
+    def _setMember(self, verb, group, member):
+        cmd = 'gerrit set-members'
+        if verb == 'add':
+            cmd = '%s --add "%s"' % (cmd, member)
+        elif verb == 'remove':
+            cmd = '%s --remove "%s"' % (cmd, member)
+        cmd = '%s "%s"' % (cmd, group)
+        out, err = self._ssh(cmd)
+        return err
+
+    def addMember(self, group, member):
+        self._setMember('add', group, member)
+
+    def removeMember(self, group, member):
+        self._setMember('remove', group, member)
+
+    def createProject(self, project, require_change_id=True, empty_repo=False,
+                      description=None):
         cmd = 'gerrit create-project'
         if require_change_id:
             cmd = '%s --require-change-id' % cmd
         if empty_repo:
             cmd = '%s --empty-commit' % cmd
-        cmd = '%s --name %s' % (cmd, project)
+        if description:
+            cmd = "%s --description \"%s\"" % \
+                  (cmd, description.replace('"', r'\"'))
+        version = None
+        try:
+            version = self.parseGerritVersion(self.getVersion())
+        except Exception:
+            # If no version then we know version is old and should use --name
+            pass
+        if not version or version < (2, 12):
+            cmd = '%s --name "%s"' % (cmd, project)
+        else:
+            cmd = '%s "%s"' % (cmd, project)
         out, err = self._ssh(cmd)
         return err
 
-    def listProjects(self):
+    def updateProject(self, project, update_key, update_value):
+        # check for valid keys
+        if update_key not in UPDATE_ALLOWED_KEYS:
+            raise Exception("Trying to update a non-valid key %s" % update_key)
+
+        cmd = 'gerrit set-project %s ' % project
+        if update_key == 'description':
+            cmd += "--%s \"%s\"" % (update_key,
+                                    update_value.replace('"', r'\"'))
+        else:
+            cmd += '--%s %s' % (update_key, update_value)
+        out, err = self._ssh(cmd)
+        return err
+
+    def listProjects(self, show_description=False):
         cmd = 'gerrit ls-projects'
+        if show_description:
+            # display projects alongs with descriptions
+            # separated by ' - ' sequence
+            cmd += ' --description'
         out, err = self._ssh(cmd)
         return filter(None, out.split('\n'))
 
-    def listGroups(self):
-        cmd = 'gerrit ls-groups'
+    def listGroups(self, verbose=False):
+        if verbose:
+            cmd = 'gerrit ls-groups -v'
+        else:
+            cmd = 'gerrit ls-groups'
         out, err = self._ssh(cmd)
         return filter(None, out.split('\n'))
 
@@ -245,6 +314,19 @@ class Gerrit(object):
         out = out.split(' ')[2]
         return out.strip('\n')
 
+    def parseGerritVersion(self, version):
+        # Adapted from gertty setRemoteVersion()
+        base = version.split('-')[0]
+        parts = base.split('.')
+        major = minor = micro = 0
+        if len(parts) > 0:
+            major = int(parts[0])
+        if len(parts) > 1:
+            minor = int(parts[1])
+        if len(parts) > 2:
+            micro = int(parts[2])
+        return (major, minor, micro)
+
     def replicate(self, project='--all'):
         cmd = 'replication start %s' % project
         if self.installed_plugins is None:
